#include <libavcodec/avcodec.h>
#include <libavformat/avformat.h>
#include <libavutil/avutil.h>
#include <libswscale/swscale.h>
#include <pthread.h>
#include <stdio.h>
#include "audioflinger_wrapper.h"
#include "surfaceflinger_wrapper.h"

//----------------------------------------------------------------------------------------------------
#include <utils/Log.h>
#undef  LOG_TAG
#define LOG_TAG "HoFFPLAYengin"
#if 1
#define gprintf(fmt, args...) LOGE("%s(%d): " fmt, __FUNCTION__, __LINE__, ##args)
#else
#define gprintf(fmt, args...)
#endif
//----------------------------------------------------------------------------------------------------

typedef struct PktQueue {
    AVPacketList *first_packet, *last_packet;
    int packets_cnt;
    int size;
    pthread_mutex_t mutex;
    pthread_cond_t  cond;
} PktQueue;

PktQueue    audioq; //오디오 패킷을 담을 Queue 이다.
int         quit = 0;
pthread_t   asink_tid = NULL;
AudioFlingerDeviceHandle audioflinger_device = NULL; //안드로이드의 오디오 출력을 위한 핸들

static struct SwsContext *img_convert_ctx = NULL; //Video format 변경을 위한 sw converter 세팅

//Queue를 초기화 한다.
void pkt_queue_init(PktQueue *queue) {
    memset(queue, 0, sizeof(PktQueue));
    pthread_mutex_init(&queue->mutex, NULL);
    pthread_cond_init(&queue->cond, NULL);
}

//Packet을 Queue에 넣는 부분
int pkt_queue_put(PktQueue *queue, AVPacket *pkt) {
    AVPacketList *pkt_temp;
    if(av_dup_packet(pkt) < 0) {
        return -1;
    }
    pkt_temp = av_malloc(sizeof(AVPacketList));
    if (!pkt_temp) return -1;
    pkt_temp->pkt = *pkt;
    pkt_temp->next = NULL;
    pthread_mutex_lock(&queue->mutex);
    if (!queue->last_packet)
        queue->first_packet = pkt_temp;
    else
        queue->last_packet->next = pkt_temp;
    queue->last_packet = pkt_temp;
    queue->packets_cnt++;
    queue->size += pkt_temp->pkt.size;
    pthread_cond_signal(&queue->cond);
    pthread_mutex_unlock(&queue->mutex);
    return 0;
}

//Queue의 패킷을 가져오는 부분
static int pkt_queue_get(PktQueue *q, AVPacket *pkt) {
    AVPacketList *pkt_temp;
    int ret;
    pthread_mutex_lock(&q->mutex);
    for(;;) {
        if(quit) {
            ret = -1;
            break;
        }
        pkt_temp = q->first_packet;
        if (pkt_temp) {
            q->first_packet = pkt_temp->next;
            if (!q->first_packet)
                q->last_packet = NULL;
            q->packets_cnt--;
            q->size -= pkt_temp->pkt.size;
            *pkt = pkt_temp->pkt;
            av_free(pkt_temp);
            ret = 1;
            break;
        } else {
            pthread_cond_wait(&q->cond, &q->mutex);
        }
    }
    pthread_mutex_unlock(&q->mutex);
    return ret;
}

//오디오 패킷을 디코딩 하는 함수
int decode_audio(AVCodecContext *AudioCodecCtx, uint8_t *audio_buffer) {
    static AVPacket pkt;
    static uint8_t *apkt_data = NULL;
    static int apkt_size = 0;
    int got_frame_size;
    int ret;

    for(;;) {
        while(apkt_size > 0) {

            //오디오 파일을 디코딩 한다.
            got_frame_size = AVCODEC_MAX_AUDIO_FRAME_SIZE;
            ret = avcodec_decode_audio3(AudioCodecCtx, audio_buffer, &got_frame_size, &pkt);
            if(ret < 0) {
                gprintf("avcodec_decode_audio3 fail. \n");
                apkt_size = 0;
                break;
            }
            apkt_data += ret;
            apkt_size -= ret;
            if(got_frame_size <= 0) {
                continue;
            }
            return got_frame_size;
        }
        if(pkt.data)
            av_free_packet(&pkt);
        if(quit) {
            gprintf("quit. \n");
            return -1;
        }
        //Queue 에서 오디오 패킷을 가져온다.
        if(pkt_queue_get(&audioq, &pkt) < 0) {
            gprintf("pkt_queue_get fail. \n");
            return -1;
        }
        apkt_data = pkt.data;
        apkt_size = pkt.size;
    }
}

//오디오 디코딩 콜백, 그냥 Loop를 돌면서 오디오를 디코딩한다.
void audio_task(AVCodecContext *userdata, uint8_t *stream, int len) {
    AVCodecContext *AudioCodecCtx = userdata;
    int ret, audio_size;
    static uint8_t audio_buffer[(AVCODEC_MAX_AUDIO_FRAME_SIZE * 3) / 2];
    static unsigned int audio_buffer_size = 0;
    static unsigned int audio_buffer_index = 0;
    while(len > 0) {
        if(audio_buffer_index >= audio_buffer_size) {
            audio_size = decode_audio(AudioCodecCtx, audio_buffer);
            if(audio_size < 0) {
                audio_buffer_size = 1024;
                memset(audio_buffer, 0, audio_buffer_size);
            } else {
                audio_buffer_size = audio_size;
            }
            audio_buffer_index = 0;
        }
        ret = audio_buffer_size - audio_buffer_index;
        if(ret > len)
            ret = len;
        memcpy(stream, (uint8_t *)audio_buffer + audio_buffer_index, ret);
        len -= ret;
        stream += ret;
        audio_buffer_index += ret;
    }
}

unsigned char ggg[16384];

//오디오 플레이 쓰레드
void* play(void *t) {
    int             ret=0;

    //안드로이드의 오디오 재생을 시작한다.
    audioflinger_device_start(audioflinger_device);
    for (;;) {
        //콜백을 통해 오디오 패킷을 디코딩한다.
        audio_task((AVCodecContext*)t, ggg, sizeof(ggg) );
        if( quit == 1 ) {
            gprintf("audio......quit\n");
            break;
        }
        //디코딩된 오디오 데이터를 안드로이드의 AudioTrack에 write 하여 출력한다.
        ret = audioflinger_device_write(audioflinger_device, ggg, sizeof(ggg));
        if (ret == 0) {
            gprintf("audioflinger_device_stop.... \n");
            audioflinger_device_stop(audioflinger_device);
            goto error;
        }
    }
error:
    gprintf("Audio Play End.... \n");
    pthread_exit(NULL);
    return NULL;
}

//엔진의 시작이다.
int engine_start(char* fname) {
    AVFormatContext *ptrFormatCtx = NULL;
    int             i, vStream, aStream;
    AVCodec         *vCodec;
    AVCodec         *aCodec;
    AVFrame         *vFrame;
    AVPacket        packet;
    int             endDecoding;
    AVCodecContext  *VideoCodecCtx;
    AVCodecContext  *AudioCodecCtx;
    int             ret = 0;
    AVInputFormat   *iformat = NULL;
    AVDictionary    *dict = NULL;

    av_register_all(); //ffmpeg의 모든 코덱을 등록한다.

    //동영상 파일을 연다.
    if(avformat_open_input(&ptrFormatCtx, fname, iformat, dict)!=0) {
        gprintf("Couldn't open file : %s \n", fname);
        return -1; // Couldn't open file
    }

    // 오픈한 동영상 파일의 스트림 정보를 가져온다.
    if(av_find_stream_info(ptrFormatCtx)<0) {
        gprintf("faile av_find_stream_info \n");
        return -1; // 스트림 정보를 가져오지 못한 경우
    }

    // 비디오 소스로부터 비디오 스트림과 오디오 스트림을 찾는다.
    vStream=-1;
    aStream=-1;
    for(i=0; i<ptrFormatCtx->nb_streams; i++) {
        if(ptrFormatCtx->streams[i]->codec->codec_type==AVMEDIA_TYPE_VIDEO && vStream < 0) {
            vStream=i;
        }
        if(ptrFormatCtx->streams[i]->codec->codec_type==AVMEDIA_TYPE_AUDIO && aStream < 0) {
            aStream=i;
        }
    }

    if(vStream==-1) {
        gprintf("No vStream \n");
        return -1; //비디오 스트림이 없을 경우
    }

    if(aStream==-1) {
        gprintf("No aStream \n");
        return -1; //오디오 스트림이 없을 경우
    }

    AudioCodecCtx=ptrFormatCtx->streams[aStream]->codec;
    gprintf("sample_rate : %d \n", AudioCodecCtx->sample_rate);
    gprintf("channels : %d \n", AudioCodecCtx->channels);

    //오디오 코덱 정보를 이용해서 오디오 output을 세팅한다.
    audioflinger_device = audioflinger_device_create();

    if( audioflinger_device == NULL ) {
        gprintf("audioflinger_device_create fail\n");
        exit(1);
    }

    if( audioflinger_device_set(audioflinger_device, 3, AudioCodecCtx->channels, AudioCodecCtx->sample_rate, 16384) == -1) {
        gprintf("audioflinger_device_set fail\n");
        exit(1);
    }
    gprintf("frame count: %d, frame size: %d \n",
        audioflinger_device_frameCount(audioflinger_device), audioflinger_device_frameSize(audioflinger_device) );

    audioflinger_device_pause(audioflinger_device);

    //오디오 코덱 컨텍스트(AudioCodecCtx)로부터 코덱을 찾는다.
    aCodec = avcodec_find_decoder(AudioCodecCtx->codec_id);
    if(!aCodec) {
        fprintf(stderr, "Unsupported codec!\n");
        return -1;
    }

    //오디오 코덱 오픈
    avcodec_open(AudioCodecCtx, aCodec);

    pkt_queue_init(&audioq);

    //오디오 플레이를 위한 Thread를 띄운다.
    ret = pthread_create(&asink_tid, NULL, play, (void *)AudioCodecCtx );
    if(ret) {
        gprintf("Server: Cannot create audio decode thread.\n");
        return -1;
    }

    // 오픈한 스트림 중에 비디오 스트림의 코덱 정보를 가져온다.
    VideoCodecCtx=ptrFormatCtx->streams[vStream]->codec;

    // 비디오 스트림의 코덱 id로 비디오 디코더를 찾아온다.
    vCodec=avcodec_find_decoder(VideoCodecCtx->codec_id);
    if(vCodec==NULL) {
        fprintf(stderr, "Unsupported codec!\n");
        return -1; //코덱을 찾지 못한 경우
    }

    // 비디오 코덱을 연다.
    if(avcodec_open(VideoCodecCtx, vCodec)<0)
        return -1; //코덱을 열지 못한 경우

    // 비디오 프레임을 위한 메모리 할당을 해준다.
    vFrame=avcodec_alloc_frame();

    //안드로이드 화면에 출력할 수 있도록 비디오 세팅을 한다.
    hovideo_prepare(0, 0, VideoCodecCtx->width, VideoCodecCtx->height, VideoCodecCtx->width, VideoCodecCtx->height, 21);

    //디코딩된 비디오 데이터를 화면에 보여주기 위해서 변환을 해야한다.
    //소프트웨어 컨버터를 사용하여 변환할 것이다.
    AVFrame* pFrameyuv420p = avcodec_alloc_frame();
    int destFormat = PIX_FMT_YUV420P;
    int numBytes = avpicture_get_size(destFormat, VideoCodecCtx->width, VideoCodecCtx->height);
    uint8_t* buffer = (uint8_t *)av_malloc(numBytes*sizeof(uint8_t));
    avpicture_fill((AVPicture *)pFrameyuv420p, buffer, destFormat, VideoCodecCtx->width, VideoCodecCtx->height);
    AVPicture* dst = (AVPicture*)pFrameyuv420p;

    i=0;
    av_init_packet(&packet);
    // 동영상 파일로부터 Frame을 계속 읽어들인다.
    while(av_read_frame(ptrFormatCtx, &packet)>=0) {

        // 읽어온 패킷이 비디오 스트림일 경우에는
        if(packet.stream_index==vStream) {

            // 설정한 비디오 코덱으로 디코딩을 한다.
            avcodec_decode_video2(VideoCodecCtx, vFrame, &endDecoding, &packet);

            if(endDecoding) { //비디오 프레임이 디코딩이 완료 되었다면

                //소프트웨어 스케일러로 YUV420P 타입으로 변환한다. 
                //왜냐하면 안드로이드 비디오 세팅을 YUV420P로 했기 때문이다. 만일 다른 포맷으로 했다면 거기에 맞추면 된다.
                if(img_convert_ctx == NULL) {
                    int w = VideoCodecCtx->width;
                    int h = VideoCodecCtx->height;
                    img_convert_ctx = sws_getContext(w, h, VideoCodecCtx->pix_fmt, w, h, destFormat, SWS_BICUBIC, NULL, NULL, NULL);
                    if(img_convert_ctx == NULL) {
                        gprintf("Cannot initialize the conversion context!\n");
                        exit(1);
                    }
                }
                sws_scale(img_convert_ctx, vFrame->data, vFrame->linesize, 0, VideoCodecCtx->height, dst->data, dst->linesize);

                //안드로이드 화면에 변환한 비디오를 출력한다.
                hovideo_post(dst->data[0], dst->data[1], dst->data[2]);
                av_free_packet(&packet);
            }
        } else if(packet.stream_index==aStream) { //읽어온 패킷이 오디오 스트림이라면 오디오 큐에 넣는다.
            pkt_queue_put(&audioq, &packet);
        } else { //이도 저도 아닌것이라면 버린다.
            av_free_packet(&packet);
        }
        //앱이 종료되면 멈추도록 한다.
        if(quit == 1)
            break;
    }
    gprintf("---- End ---- \n");

    // 디코딩한 프레임 메모리를 해제한다.
    av_free(vFrame);

    // 비디오 코덱을 닫는다.
    avcodec_close(VideoCodecCtx);

    // 비디오 파일을 닫는다.
    av_close_input_file(ptrFormatCtx);

    sws_freeContext(img_convert_ctx);

    //안드로이드의 비디오와 오디오를 종료한다.
    hovideo_release();
    audioflinger_device_stop(audioflinger_device);
    audioflinger_device_release(audioflinger_device);
    return 0;
}

